/*

Sim67 is a ROM based emulator for HP-67
(Copyright 2007-2013) Jacques LAPORTE Version 2.6
Build Feb 12 2013
Portions Copyright Eric Smith 1995-2009 & David G. Hicks (Java Port)

*/

import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.*;
import java.awt.Font;

import java.awt.FontFormatException;
import java.io.IOException;
import java.io.InputStream;

public class sim67 extends Applet {

	// ////////////////////////////////////////////////////////////////////////
	//
	// The following methods display the UI and
	// handle the mouse actions
	//
	// ////////////////////////////////////////////////////////////////////////

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	Font key_font, disp_font, trace_font;
	FontMetrics kfm, dfm, tfm;
	Graphics gr;
	String disp_buf = new String("");

	Frame debugframe;
	TextArea debugtext;

	Frame cardframe;
	TextArea cardtext;

	int pressed_key = -1;
	int pressed_debug_key = -1;

	boolean debug_mode = true;

	Image img = null;
	Image himg = null;

	static final int PGM_MODE = 0001;
	static final int RUN_MODE = 0002;
	static final int OFF_MODE = 0003;
	static final int ON_MODE = 0004;

	static final int WINDOW_WIDTH = 254;
	static final int WINDOW_HEIGHT = 584;
	static final int DISPLAY_HEIGHT = 52;

	Rectangle display_rect = new Rectangle(0, 0, WINDOW_WIDTH, DISPLAY_HEIGHT);

	static final int BLACK = 0;
	static final int WHITE = 1;
	static final int LT_GREY = 2;
	static final int MED_GREY = 3;
	static final int DK_GREY = 4;
	static final int GOLD = 5;
	static final int BLUE = 6;
	static final int DK_RED = 7;
	static final int BR_RED = 8;
	static final int VLT_GREY = 9;
	static final int GREEN = 10;

	Color color[] = { new Color(0x00, 0x00, 0x00), new Color(0xff, 0xff, 0xff),
			new Color(0xa0, 0xa0, 0xa0), new Color(0x58, 0x58, 0x58),
			new Color(0x38, 0x38, 0x38), new Color(0xff, 0xd7, 0x00),
			new Color(0x40, 0x40, 0xff), new Color(0x50, 0x00, 0x00),
			new Color(0xff, 0x40, 0x00), new Color(0xc0, 0xc8, 0xd8),
			new Color(0x85, 0x80, 0x09)

	};

	class Keyinfo {
		Rectangle rect;
		String label;
		String flabel;
		String glabel;
		String hlabel;
		int keycode;
		int fg, bg;

		// Rectangle(int x, int y, int width, int height) : x,y upper left
		// corner coordinates

		public Keyinfo(Rectangle rect, String label, String flabel,
				String glabel, String hlabel, int keycode, int fg, int bg) {
			this.rect = rect;
			this.label = label;
			this.flabel = flabel;
			this.glabel = glabel;
			this.hlabel = hlabel;
			this.keycode = keycode;
			this.fg = fg;
			this.bg = bg;
		}

	};

	static final int LABEL = 85;
	static final int A = LABEL + 40;
	static final int B = A + 60;
	static final int C = B + 60;
	static final int D = C + 60;
	static final int E = D + 60;
	static final int F = E + 60;
	static final int G = F + 60;

	Keyinfo keys[] = {
			new Keyinfo(new Rectangle(145, 48, 60, 20), "W/PRGM", "", "", "",
					PGM_MODE, WHITE, BLACK),
			new Keyinfo(new Rectangle(206, 48, 30, 20), "RUN", "", "", "",
					RUN_MODE, WHITE, BLACK),
			new Keyinfo(new Rectangle(15, 55, 30, 20), "    ", "", "", "",
					OFF_MODE, WHITE, BLACK),
			new Keyinfo(new Rectangle(60, 55, 30, 20), "    ", "", "", "",
					ON_MODE, WHITE, BLACK),

			new Keyinfo(new Rectangle(14, LABEL, 30, 20), "A", "a", "", "",
					0244, WHITE, GREEN),
			new Keyinfo(new Rectangle(62, LABEL, 30, 20), "B", "b", "", "",
					0243, WHITE, GREEN),
			new Keyinfo(new Rectangle(110, LABEL, 30, 20), "C", "c", "", "",
					0242, WHITE, GREEN),
			new Keyinfo(new Rectangle(158, LABEL, 30, 20), "D", "d", "", "",
					0241, WHITE, GREEN),
			new Keyinfo(new Rectangle(206, LABEL, 30, 20), "E", "e", "", "",
					0240, WHITE, GREEN),

			new Keyinfo(new Rectangle(14, A, 30, 20), "S+", "x", "s", "", 0224,
					WHITE, GREEN),
			new Keyinfo(new Rectangle(62, A, 30, 20), "GTO", "GSB", "f", "",
					0223, WHITE, GREEN),
			new Keyinfo(new Rectangle(110, A, 30, 20), "DSP", "FIX", "SCI", "",
					0222, WHITE, GREEN),
			new Keyinfo(new Rectangle(158, A, 30, 20), "(i)", "RND", "", "",
					0221, WHITE, GREEN),
			new Keyinfo(new Rectangle(206, A, 30, 20), "SST", "LBL", "f", "",
					0220, WHITE, GREEN),

			new Keyinfo(new Rectangle(14, B, 30, 20), "f", "", "", "", 0024,
					BLACK, GOLD),
			new Keyinfo(new Rectangle(62, B, 30, 20), "g", "", "", "", 0023,
					BLACK, BLUE),
			new Keyinfo(new Rectangle(110, B, 30, 20), "STO", "DSZ", "(i)", "",
					0022, WHITE, GREEN),
			new Keyinfo(new Rectangle(158, B, 30, 20), "RCL", "ISZ", "(i)", "",
					0021, WHITE, GREEN),
			new Keyinfo(new Rectangle(206, B, 30, 20), "h", "", "", "", 0020,
					WHITE, BLACK),

			new Keyinfo(new Rectangle(14, C, 78, 20), "ENTER^", "W/DATA",
					"MERGE", "", 0063, WHITE, GREEN),
			new Keyinfo(new Rectangle(110, C, 30, 20), "CHS", "P<>S", "", "",
					0062, WHITE, GREEN),
			new Keyinfo(new Rectangle(158, C, 30, 20), "EEX", "CL REG", "", "",
					0061, WHITE, GREEN),
			new Keyinfo(new Rectangle(206, C, 30, 20), "CLX", "CL PGM", "", "",
					0060, WHITE, GREEN),

			new Keyinfo(new Rectangle(14, D, 24, 20), "-", "x=0", "x=y", "SF",
					0103, WHITE, GREEN),
			new Keyinfo(new Rectangle(63, D, 37, 20), "7", "ln", "e^x", "x><y",
					0102, BLACK, WHITE),
			new Keyinfo(new Rectangle(131, D, 37, 20), "8", "log", "10^x",
					"Rv", 0101, BLACK, WHITE),
			new Keyinfo(new Rectangle(199, D, 37, 20), "9", "Vx", "x^2", "R^",
					0100, BLACK, WHITE),

			new Keyinfo(new Rectangle(14, E, 24, 20), "+", "x#0", "x#y", "CF",
					0123, WHITE, GREEN),
			new Keyinfo(new Rectangle(63, E, 37, 20), "4", "sin", "sin-1",
					"1/x", 0122, BLACK, WHITE),
			new Keyinfo(new Rectangle(131, E, 37, 20), "5", "cos", "cos-1",
					"y^x", 0121, BLACK, WHITE),
			new Keyinfo(new Rectangle(199, E, 37, 20), "6", "tan", "tan-1",
					"ABS", 0120, BLACK, WHITE),

			new Keyinfo(new Rectangle(14, F, 24, 20), "x", "x<0", "x<=y", "F?",
					0143, WHITE, GREEN),
			new Keyinfo(new Rectangle(63, F, 37, 20), "1", "R<", ">P", "PAUSE",
					0142, BLACK, WHITE),
			new Keyinfo(new Rectangle(131, F, 37, 20), "2", "D<", ">R", "Pi",
					0141, BLACK, WHITE),
			new Keyinfo(new Rectangle(199, F, 37, 20), "3", "H<", ">HMS",
					"REG", 0140, BLACK, WHITE),

			new Keyinfo(new Rectangle(14, G, 24, 20), "/", "x>0", "x>y", "n!",
					0163, WHITE, GREEN),
			new Keyinfo(new Rectangle(63, G, 37, 20), "0", "%", "%CH", "LSTx",
					0162, BLACK, WHITE),
			new Keyinfo(new Rectangle(131, G, 37, 20), ".", "INT", "FRAC",
					"H.MS+", 0161, BLACK, WHITE),
			new Keyinfo(new Rectangle(199, G, 37, 20), "R/S", "-x-", "STK",
					"SPACE", 0160, BLACK, WHITE)

	};

	class DebugKeyinfo {
		Rectangle rect;
		String label;
		int fg, bg;
		boolean isKey; // if false then this is just a label
		boolean toggle;
		boolean value;

		public DebugKeyinfo(Rectangle rect, String label, int fg, int bg,
				boolean isKey) {
			this.rect = rect;
			this.label = label;
			this.fg = fg;
			this.bg = bg;
			this.isKey = isKey;
			toggle = false;
			value = false;
		}

		public DebugKeyinfo(Rectangle rect, String label, int fg, int bg,
				boolean isKey, boolean toggle, boolean value) {
			this.rect = rect;
			this.label = label;
			this.fg = fg;
			this.bg = bg;
			this.isKey = isKey;
			this.toggle = toggle;
			this.value = value;
		}
	};

	final static int LEDSIZE = 6;

	final static int TRACE = 0; // constants must be in the same order
	final static int REGS = 1; // as the dkeys array below
	final static int RUN = 2;
	final static int SLOW = 3;
	final static int STEP = 4;
	final static int STOP = 5;
	final static int SET1 = 6;
	final static int SET2 = 7;
	final static int SET3 = 8;
	final static int ENABLE1 = 9;
	final static int ENABLE2 = 10;
	final static int ENABLE3 = 11;
	final static int BP0 = 12;
	final static int BP1 = 13;
	final static int BP2 = 14;
	final static int BP3 = 15;
	final static int BP4 = 16;
	final static int BP5 = 17;
	final static int BP6 = 18;
	final static int BP7 = 19;
	final static int BPCLR = 20;
	final static int DEBUG = 21;
	final static int CARD = 22;
	final static int MEMP = 23;
	final static int MEMD = 24;
	final static int CARD2 = 25;
	final static int CRDCLR = 26;
	final static int MEMS = 27;

	DebugKeyinfo dkeys[] = {
			new DebugKeyinfo(new Rectangle(270, 199, 70, 24), "    trace",
					WHITE, MED_GREY, true, true, false),

			new DebugKeyinfo(new Rectangle(270, 229, 70, 24), "    regs",
					WHITE, MED_GREY, true, true, false),

			new DebugKeyinfo(new Rectangle(270, 62, 70, 24), "    ON", WHITE,
					MED_GREY, true, true, false),
			new DebugKeyinfo(new Rectangle(270, 92, 70, 24), "    slow", WHITE,
					MED_GREY, true, true, false),
			new DebugKeyinfo(new Rectangle(270, 122, 70, 24), "    step",
					WHITE, MED_GREY, true),
			new DebugKeyinfo(new Rectangle(270, 152, 70, 24), "   OFF", WHITE,
					MED_GREY, true),

			// breakpoint set and enable keys
			new DebugKeyinfo(new Rectangle(540, 25, 33, 24), "set", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(540, 55, 33, 24), "set", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(540, 85, 33, 24), "set", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(680, 25, 64, 24), "    enable",
					WHITE, MED_GREY, true, true, false),
			new DebugKeyinfo(new Rectangle(680, 55, 64, 24), "    enable",
					WHITE, MED_GREY, true, true, false),
			new DebugKeyinfo(new Rectangle(680, 85, 64, 24), "    enable",
					WHITE, MED_GREY, true, true, false),

			// breakpoint digit keys
			new DebugKeyinfo(new Rectangle(540, 154, 28, 24), "0", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(590, 154, 28, 24), "1", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(630, 154, 28, 24), "2", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(670, 154, 28, 24), "3", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(540, 119, 28, 24), "4", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(590, 119, 28, 24), "5", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(630, 119, 28, 24), "6", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(670, 119, 28, 24), "7", WHITE,
					MED_GREY, true),

			new DebugKeyinfo(new Rectangle(670, 199, 28, 24), "clr", WHITE,
					MED_GREY, true),

			new DebugKeyinfo(new Rectangle(270, 12, 70, 24), "   debug", WHITE,
					MED_GREY, true, true, true),
			new DebugKeyinfo(new Rectangle(710, 234, 64, 24), "CARD <", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(710, 269, 28, 24), "MP", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(670, 269, 28, 24), "MD", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(245, 70, 10, 15), "<", WHITE, GREEN,
					true),

			new DebugKeyinfo(new Rectangle(710, 199, 28, 24), "clrCd", WHITE,
					MED_GREY, true),
			new DebugKeyinfo(new Rectangle(630, 269, 28, 24), "Stat", WHITE,
					MED_GREY, true),

			new DebugKeyinfo(new Rectangle(510, 1, 170, 24), "breakpoints",
					BLACK, VLT_GREY, true),
			new DebugKeyinfo(new Rectangle(270, 370, 465, 24), "PC", BLACK,
					VLT_GREY, true),
			new DebugKeyinfo(new Rectangle(375, 2, 100, 24),
					"internal registers", BLACK, VLT_GREY, true)

	};

	synchronized void draw_toggle_key(Graphics pgr, String s, Rectangle rect,
			int fg, int bg, boolean toggle) {

		draw_string(pgr, s, rect, fg, bg, kfm);

		if (toggle)
			pgr.setColor(color[BR_RED]);
		else
			pgr.setColor(color[DK_RED]);

		pgr.fillOval(rect.x + LEDSIZE / 2 + 3, rect.y + rect.height / 2
				- LEDSIZE / 2, LEDSIZE, LEDSIZE);
	}

	void update_run_key() {
		draw_toggle_key(gr, dkeys[RUN].label, dkeys[RUN].rect, dkeys[RUN].fg,
				dkeys[RUN].bg, dkeys[RUN].value);
	}

	void handle_debug_key(int key) {

		// enable only the DEBUG key when the debugger is off
		if (!dkeys[DEBUG].value && (key != DEBUG) && (key != CARD2))
			return;

		switch (key) {
		case TRACE:
		case REGS:
		case RUN:
		case ENABLE1:
		case ENABLE2:
		case ENABLE3:
			dkeys[key].value = !dkeys[key].value;
			anyBPsEnabled = dkeys[ENABLE1].value || dkeys[ENABLE2].value
					|| dkeys[ENABLE3].value;
			break;
		case BP0:
		case BP1:
		case BP2:
		case BP3:
		case BP4:
		case BP5:
		case BP6:
		case BP7:
			if (key > 11 && key < 20) {
				BpIn = BpIn * 10;

				BpIn += key - BP0;
				display_bpzone();

			}
			bpinput <<= 3;
			bpinput += key - BP0;

			break;
		case BPCLR:
			bpinput = 0;
			BpIn = 0;
			display_bpzone();
			break;
		case CRDCLR:
			cardframe.setVisible(true);
			cardtext.setText("");
			debugtext.setText("");
			debugframe.setVisible(true);
			break;
		case DEBUG:
			dkeys[key].value = !dkeys[key].value;
			if (!dkeys[key].value) {
				debugframe.setVisible(false);
				debug_mode = false;
			} else {
				debugframe.setVisible(true);
				debug_mode = true;
			}
			if (!dkeys[key].value) {
				// debugger turned off
				dkeys[SLOW].value = false;
				dkeys[RUN].value = true;
				step = false;
				dkeys[ENABLE1].value = dkeys[ENABLE2].value = dkeys[ENABLE3].value = false;
				anyBPsEnabled = false;
				dkeys[REGS].value = false;
				dkeys[TRACE].value = false;
			}
			draw_debugger(gr);

			break;
		case SET1:
		case SET2:
		case SET3:
			dkeys[key].value = !dkeys[key].value;
			bpReg[key - SET1] = bpinput;
			bpinput = 0;
			display_user_bps();

			BpIn = 0;
			display_bpzone();

			break;
		case SLOW:
			dkeys[key].value = !dkeys[key].value;
			if (dkeys[key].value) {
				dkeys[RUN].value = true;
				update_run_key();
			}
			break;
		case CARD:
		case CARD2:

			if (!dkeys[RUN].value)
				break;
			content = cardtext.getText();

			int len;
			len = content.length();
			cardframe.setVisible(true);

			if (prgm_mode == false) { // RUN
				if (len > 35)
					card_inserted = 1; // pull the card to READ if a card is
										// there
				else // if RUN mode with an empty card
				if (write_card == 1 && len == 0)
					card_inserted = 1; // Write data
				else // error and no read
				{
					debug("Invalid Card image length " + Integer.toString(len));
					card_inserted = 0;
				}
			} else
				// PRGM
				card_inserted = 1; // pull the card to Write a PRGM

			break;
		case MEMP:
			display_mem_pgm();
			break;
		case MEMD:
			display_mem_data(bpinput);
			break;
		case MEMS:

			display_status_data();
			break;

		case STEP:
			dkeys[RUN].value = false;
			update_run_key();
			step = true;
			break;
		case STOP:
			dkeys[RUN].value = false;
			update_run_key();
			break;
		}
		if (dkeys[key].toggle) {
			draw_toggle_key(gr, dkeys[key].label, dkeys[key].rect,
					dkeys[key].fg, dkeys[key].bg, dkeys[key].value);
		}

	}

	synchronized void draw_string(Graphics pgr, String s, Rectangle rect,
			int fg, int bg, FontMetrics fm) {
		int descent;
		int sheight, swidth;

		if (gr == null)
			System.out.println("gr is null!!!");
		pgr.setColor(color[bg]);
		pgr.fillRect(rect.x, rect.y, rect.width, rect.height);

		if (s != null) {
			if (!s.equals("")) {
				pgr.setColor(color[fg]);
				descent = fm.getMaxDescent();
				sheight = fm.getHeight();
				swidth = fm.stringWidth(s);
				pgr.drawString(s, rect.x + (rect.width - swidth) / 2, rect.y
						+ (rect.height * 2 - sheight) / 2 + descent);

			}
		}
	}

	synchronized void draw_string_left(Graphics pgr, String s, Rectangle rect,
			int fg, int bg, FontMetrics fm) {
		int descent;
		int sheight;

		if (pgr == null)
			System.out.println("pgr is null!!!");

		pgr.setColor(color[bg]);
		pgr.fillRect(rect.x, rect.y, rect.width, rect.height);

		if (s != null) {
			if (!s.equals("")) {
				pgr.setColor(color[fg]);
				descent = fm.getMaxDescent();
				sheight = fm.getHeight();
				pgr.drawString(s, rect.x + 2, rect.y
						+ (rect.height * 2 - sheight) / 2 + descent);

			}
		}
	}

	synchronized void draw_display(Graphics pgr, String s) {
		pgr.setFont(disp_font);
		draw_string(pgr, s, display_rect, BR_RED, DK_RED, dfm);
		pgr.setFont(key_font);
	}

	synchronized void draw_calc(Graphics pgr) {

		String S = "hp67.gif";
		if (S != null) {
			MediaTracker m = new MediaTracker(this);
			img = getImage(getCodeBase(), S);
			m.addImage(img, 0);
			try {
				m.waitForID(0);
			} catch (InterruptedException e) {
			}
		}

		String Sh = "hp67h.gif";
		if (Sh != null) {
			MediaTracker m2 = new MediaTracker(this);
			himg = getImage(getCodeBase(), Sh);
			m2.addImage(himg, 0);
			try {
				m2.waitForID(0);
			} catch (InterruptedException e) {
			}
		}

		gr.drawImage(img, 0, 0, this);

	}

	synchronized void draw_basic_debugger(Graphics pgr) {
		int i;
		// Draw debugger when OFF and change background color
		int debugcolor;
		if (debug_mode)
			debugcolor = VLT_GREY;
		else
			debugcolor = WHITE;
		pgr.setColor(color[debugcolor]);
		pgr.fillRect(WINDOW_WIDTH, 0, 760 - WINDOW_WIDTH, WINDOW_HEIGHT);

		if (dkeys[DEBUG].value) {
			for (i = 0; i < dkeys.length; i++) {
				if (dkeys[i].toggle)
					draw_toggle_key(pgr, dkeys[i].label, dkeys[i].rect,
							dkeys[i].fg, dkeys[i].bg, dkeys[i].value);
				else
					draw_string(pgr, dkeys[i].label, dkeys[i].rect,
							dkeys[i].fg, dkeys[i].bg, kfm);
			}
		} else {
			draw_toggle_key(pgr, dkeys[DEBUG].label, dkeys[DEBUG].rect,
					dkeys[DEBUG].fg, dkeys[DEBUG].bg, dkeys[DEBUG].value);
		}
	}

	// used for font calculations
	int buttonWidth = 30, buttonHeight = 24;

	synchronized public void init() {
		int i;

		System.out.println("Start Class init");
		gr = getGraphics();

		// loop until the font is small enough to fit the register value
		for (i = 24; i > 6; i--) {
			key_font = new Font("Helvetica", Font.PLAIN, i);
			kfm = getFontMetrics(key_font);

			if ((buttonWidth >= (kfm.stringWidth("x<>y") + 4))
					&& (buttonHeight >= (kfm.getHeight() + 2))
					&& (DISPLAYREGWIDTH >= (kfm.stringWidth(" 01234567890123") + 2)))
				break;
		}

		init_trace();
		init_next();
		init_reg_display();
		init_user_bp_display();

		try {

			System.out.println("Before loading font");
			InputStream in = getClass().getResourceAsStream("LED-67.ttf");
			Font font = Font.createFont(Font.TRUETYPE_FONT, in);
			disp_font = font.deriveFont(20f);
		}

		catch (IOException e1) {
			System.out.println("Font TTF e1");
		}

		catch (FontFormatException e2) {
			System.out.println("Font TTF e2");
		}

		dfm = getFontMetrics(disp_font);
		gr.setFont(key_font); // this line causes infinite repaints if in paint

		if (!calcRunning) {
			calcRunning = true;
			System.out.println("Starting the calc thread");
			calc_thread = new Thread(new run_calc());
			calc_thread.start();
		}

	}

	public void stop_calc() {
		if (calcRunning) {
			calc_thread.stop();
			my_sleep(500);
		}
	}

	public void destroy() {
		stop_calc();
		gr.dispose();
	}

	Thread calc_thread;

	synchronized public void paint(Graphics pgr) {
		draw_calc(pgr);
		draw_display(pgr, disp_buf);
		draw_debugger(pgr);
	}

	synchronized public void draw_debugger(Graphics pgr) {
		draw_basic_debugger(pgr);
		if (dkeys[DEBUG].value) {
			display_trace();
			display_next();
			display_regs(true);
			display_user_bps();
		}
	}

	boolean pt_in_rect(int x, int y, Rectangle rect) {
		return ((x >= rect.x) && (x < (rect.x + rect.width)) && (y >= rect.y) && (y < (rect.y + rect.height)));
	}

	int find_key(int x, int y) {
		int i;

		for (i = 0; i < keys.length; i++) {
			if (pt_in_rect(x, y, keys[i].rect))
				return (i);
		}
		return (-1);
	}

	int find_debug_key(int x, int y) {
		int i;

		for (i = 0; i < dkeys.length; i++) {
			if (dkeys[i].isKey) {
				if (pt_in_rect(x, y, dkeys[i].rect))
					return (i);
			}
		}
		return (-1);
	}

	public boolean mouseDown(Event e, int x, int y) {

		int key = find_key(x, y);

		if (key >= 0) {
			pressed_key = key;
		} else {
			key = find_debug_key(x, y);
			if (key >= 0) {
				pressed_debug_key = key;
			}
		}
		return true;
	}

	public boolean mouseUp(Event e, int x, int y) {
		return true;
	}

	void update_display(String str) {
		if (str != null) {
			if (!disp_buf.equals(str)) {
				disp_buf = new String(str);
				draw_display(gr, str);
			}
		}
	}

	void check_debug_keyboard() {
		if (pressed_debug_key >= 0) {
			handle_debug_key(pressed_debug_key);
			pressed_debug_key = -1;
		}
	}

	/* returns -1 if no key pressed */
	int check_keyboard() {
		check_debug_keyboard();

		if (pressed_key >= 0) {
			return (keys[pressed_key].keycode);
		} else
			return (-1);
	}

	// ////////////////////////////////////////////////////////////////////////
	//
	// The following methods emulate the ACT
	//
	// ////////////////////////////////////////////////////////////////////////

	Operation do_op_clear_reg = null;
	Operation do_op_clear_s = null;

	public interface Operation {
		public void execute(int opcode);
	}

	static final int WSIZE = 14;

	byte a[] = new byte[WSIZE];
	byte b[] = new byte[WSIZE];
	byte c[] = new byte[WSIZE];
	byte y[] = new byte[WSIZE];
	byte z[] = new byte[WSIZE];
	byte t[] = new byte[WSIZE];
	byte m1[] = new byte[WSIZE];
	byte m2[] = new byte[WSIZE];

	byte tmp[] = new byte[1];

	int p;
	int p_change[] = new int[3];
	int val;

	byte f;
	int arithmetic_base;

	static final int MAX_RAM = 160; // Max Ram = 0xa0
	int max_ram = MAX_RAM;
	int ram_addr;

	byte ram[][] = new byte[MAX_RAM][WSIZE];
	static final int SSIZE = 16;

	static final int STACK_SIZE = 2;

	byte[] s = new byte[SSIZE];
	byte[] ext_flag = new byte[SSIZE];

	byte carry, prev_carry;

	int pc;
	int rom;
	int group;

	boolean del_rom_flag;
	boolean if_flag;
	boolean prev_if_flag;

	boolean f1_flag;

	int del_rom;
	int del_grp;

	int bank;
	int ret_pc;

	int sp;
	int[] return_stack = new int[STACK_SIZE];

	boolean display_enable;
	boolean dispaly_14_digit;

	boolean def_fn = false;
	boolean merge = false;
	boolean pause = false;
	boolean a_key = false;

	boolean display_digits = false;

	int flag_count = 0;

	int record = 0;
	int block = 0;

	int left_scan;
	int right_scan;
	int display_scan_position;
	int display_digit_position;
	boolean display_14_digit;

	boolean black;

	boolean key_flag;
	boolean prgm_mode;
	boolean prev_key_flag;

	int key_buf;
	int io_count;

	int prev_pc;

	static final int MAX_GROUP = 2; // no used in Woodstock
	static final int MAX_ROM = 8; // not used in WS
	static final int ROM_SIZE = 6200; // 5120 words LIMIT ROM = 13777+1 = 6144

	// ROM MAP in emulator
	// 0000-1777
	// 2000-3777
	// 4000-5777
	// 6000-7777
	// 12000-13777
	// it takes 6144 words instead of 5120, BUT addressing is simpler (gto 12000
	// exists!).

	short ucode[][][] = new short[MAX_GROUP][MAX_ROM][ROM_SIZE];
	// break points for unintiallized ROM
	byte bpt[][][] = new byte[MAX_GROUP][MAX_ROM][ROM_SIZE];
	String source[][][] = new String[MAX_GROUP][MAX_ROM][ROM_SIZE];

	boolean step = false;
	boolean keep_src = true;

	void fatal(String s) {
		System.out.println("fatal error: " + s);
		stop_calc();
		this.destroy();
	}

	byte do_add(byte x, byte y) {
		int res;

		res = x + y + carry;
		if (res >= arithmetic_base) {
			res -= arithmetic_base;
			carry = 1;
		} else
			carry = 0;
		return (byte) res;
	}

	byte do_sub(byte x, byte y) {
		int res;

		res = (x - y) - carry;
		if (res < 0) {
			res += arithmetic_base;
			carry = 1;
		} else
			carry = 0;
		return (byte) res;
	}

	//

	class bad_op implements Operation {
		public void execute(int opcode) {
			System.out.println("illegal opcode " + opcode + " at " + prev_pc);
		}
	}

	class op_arith implements Operation {
		public void execute(int opcode) {
			int op, field;
			
			int first = 0, last = 0;
			;
			byte temp;
			int i;
			byte t[] = new byte[WSIZE];

			op = opcode >> 5;
			field = (opcode >> 2) & 7;

			switch (field) {
			case 0: /* p */
				first = p;
				last = p;

				if (p >= WSIZE) {
					System.out.println("Warning! p > WSIZE at " + prev_pc);
					last = 0; /* don't do anything */
				}
				break;
			case 1: /* wp */
				first = 0;
				last = p;
				if (p > 13) {
					System.out.println("Warning! p >= WSIZE at " + prev_pc);
					last = 13;
				}
				break;
			case 2: /* xs */
				first = 2;
				last = 2;
				break;
			case 3: /* x */
				first = 0;
				last = 2;
				break;
			case 4: /* s */
				first = 13;
				last = 13;
				break;
			case 5: /* m */
				first = 3;
				last = 12;
				break;
			case 6: /* w */
				first = 0;
				last = 13;
				break;
			case 7: /* ms */
				first = 3;
				last = 13;
				break;
			}

			switch (op) {
			case 0x00: /* 0 -> a[f] */

				for (i = first; i <= last; i++)
					a[i] = 0;
				carry = 0;
				break;

			case 0x01: /* 0 -> b[f] */
				for (i = first; i <= last; i++)
					b[i] = 0;
				carry = 0;
				break;

			case 0x02: /* a exchange b[f] */

				for (i = first; i <= last; i++) {
					temp = a[i];
					a[i] = b[i];
					b[i] = temp;
				}
				carry = 0;
				break;

			case 0x03: /* a -> b[f] */
				for (i = first; i <= last; i++)
					b[i] = a[i];
				carry = 0;
				break;

			case 0x04: /* a exchange c[f] */
				for (i = first; i <= last; i++) {
					temp = a[i];
					a[i] = c[i];
					c[i] = temp;
				}
				carry = 0;
				break;

			case 0x05: /* c -> a[f] */
				for (i = first; i <= last; i++)
					a[i] = c[i];
				carry = 0;
				break;

			case 0x06: /* b -> c[f] */
				for (i = first; i <= last; i++)
					c[i] = b[i];
				carry = 0;
				break;

			case 0x07: /* b exchange c[f] */
				for (i = first; i <= last; i++) {
					temp = b[i];
					b[i] = c[i];
					c[i] = temp;
				}
				carry = 0;
				break;

			case 0x08: /* 0 -> c[f] */
				for (i = first; i <= last; i++)
					c[i] = 0;
				carry = 0;
				break;

			case 0x09: /* a + b -> a[f] */
				carry = 0;
				for (i = first; i <= last; i++)
					a[i] = do_add(a[i], b[i]);
				break;

			case 0x0a: /* a + c -> a[f] */
				carry = 0;
				for (i = first; i <= last; i++)
					a[i] = do_add(a[i], c[i]);
				break;

			case 0x0b: /* c + c -> c[f] */
				carry = 0;
				for (i = first; i <= last; i++)
					c[i] = do_add(c[i], c[i]);
				break;

			case 0x0c: /* a + c -> c[f] */
				carry = 0;
				for (i = first; i <= last; i++)
					c[i] = do_add(a[i], c[i]);
				break;

			case 0x0d: /* a + 1 -> a[f] */
				carry = 1;
				temp = 0;
				for (i = first; i <= last; i++)
					a[i] = do_add(a[i], temp);
				break;

			case 0x0e: /* shift left a[f] */
				for (i = last; i >= first; i--)
					a[i] = (i == first) ? 0 : a[i - 1];
				carry = 0;
				break;

			case 0x0f: /* c + 1 -> c[f] */
				carry = 1;
				temp = 0;
				for (i = first; i <= last; i++)
					c[i] = do_add(c[i], temp);
				break;

			case 0x10: /* a - b -> a[f] */
				carry = 0;
				for (i = first; i <= last; i++)
					a[i] = do_sub(a[i], b[i]);
				break;

			case 0x11: /* a - c -> c[f] */
				carry = 0;
				for (i = first; i <= last; i++)
					c[i] = do_sub(a[i], c[i]);
				break;

			case 0x12: /* a - 1 -> a[f] */
				carry = 1;
				temp = 0;
				for (i = first; i <= last; i++)
					a[i] = do_sub(a[i], temp);
				break;

			case 0x13: /* c - 1 -> c[f] */
				carry = 1;
				temp = 0;
				for (i = first; i <= last; i++)
					c[i] = do_sub(c[i], temp);
				break;

			case 0x14: /* 0 - c -> c[f] */
				carry = 0;
				temp = 0;
				for (i = first; i <= last; i++)
					c[i] = do_sub(temp, c[i]);
				break;

			case 0x15: /* 0 - c - 1 -> c[f] */
				carry = 1;
				temp = 0;
				for (i = first; i <= last; i++)
					c[i] = do_sub(temp, c[i]);
				break;

			case 0x16: /* if b[f] = 0 */
				if_flag = true;
				temp = 0;

				for (i = first; i <= last; i++)
					carry |= (b[i] != 0) ? 1 : 0;

				break;

			case 0x17: /* if c[f] = 0 */
				if_flag = true;
				temp = 0;
				for (i = first; i <= last; i++)
					carry |= (c[i] != 0) ? 1 : 0;

				break;

			case 0x18: /* if a >= c[f] */
				if_flag = true;
				carry = 0;
				for (i = first; i <= last; i++)
					t[i] = do_sub(a[i], c[i]);
				break;
			case 0x19: /* if a >= b[f] */
				if_flag = true;
				carry = 0;
				for (i = first; i <= last; i++)
					t[i] = do_sub(a[i], b[i]);
				break;

			case 0x1a: /* if a[f] # 0 */
				if_flag = true;
				carry = 1;

				for (i = first; i <= last; i++)
					carry &= (a[i] == 0) ? 1 : 0;
				break;

			case 0x1b: /* if c[f] # 0 */
				if_flag = true;
				carry = 1;

				for (i = first; i <= last; i++)
					carry &= (c[i] == 0) ? 1 : 0;
				break;

			case 0x1c: /* a - c -> a[f] */
				carry = 0;
				for (i = first; i <= last; i++)
					a[i] = do_sub(a[i], c[i]);
				break;
			case 0x1d: /* shift right a[f] */
				for (i = first; i <= last; i++)
					a[i] = (i == last) ? 0 : a[i + 1];
				carry = 0;
				break;
			case 0x1e: /* shift right b[f] */
				for (i = first; i <= last; i++)
					b[i] = (i == last) ? 0 : b[i + 1];
				carry = 0;
				break;
			case 0x1f: /* shift right c[f] */
				for (i = first; i <= last; i++)
					c[i] = (i == last) ? 0 : c[i + 1];
				carry = 0;
				break;

			}
		}
	}

	class op_goto implements Operation {
		public void execute(int opcode) {
			if (prev_carry == 0) {

				pc = (pc & ~0377) | (opcode >> 2);
				if (del_rom_flag) {
					pc = (del_rom << 8) + (pc & 0377);
					del_rom_flag = false;
				}

			}
		}
	}

	class op_jsb implements Operation {
		public void execute(int opcode) {

			return_stack[sp] = pc;
			sp++;
			if (sp >= STACK_SIZE) {
				// System.out.println("stack overflow\n");
				sp = 0;
			}
			pc = (pc & ~0377) | (opcode >> 2);
			if (del_rom_flag) {

				pc = (del_rom << 8) + (pc & 0377);
				del_rom_flag = false;
				

			}

		}
	}

	class op_return implements Operation {
		public void execute(int opcode) {

			sp--;
			if (sp < 0) {
				// System.out.println("stack underflow\n");
				sp = STACK_SIZE - 1;
			}
			pc = return_stack[sp];

		}
	}

	class op_nop implements Operation {
		public void execute(int opcode) {
		}
	}

	class op_binary implements Operation {
		public void execute(int opcode) {
			arithmetic_base = 16;
		}
	}

	class op_decimal implements Operation {
		public void execute(int opcode) {
			arithmetic_base = 10;
		}
	}

	class op_dec_p implements Operation {
		public void execute(int opcode) {

			p_change[0] = -1;
			if (p != 0)
				p--;
			else
				p = WSIZE - 1;
			// p = (p - 1) & 0xf;
		}
	}

	class op_inc_p implements Operation {
		public void execute(int opcode) {

			p_change[0] = 1;
			p++;
			if (p >= WSIZE)
				p = 0;
		}
	}

	class op_clear_s implements Operation {
		public void execute(int opcode) {
			int i;
			for (i = 0; i < SSIZE; i++)
				if ((i != 1) && (i != 2) && (i != 5) && (i != 15))
					s[i] = 0;

		}
	}

	class op_m1_exch_c implements Operation {
		public void execute(int opcode) {
			int i;
			byte t;
			for (i = 0; i < WSIZE; i++) {
				t = c[i];
				c[i] = m1[i];
				m1[i] = t;
			}
		}
	}

	class op_m1_to_c implements Operation {
		public void execute(int opcode) {
			int i;
			for (i = 0; i < WSIZE; i++)
				c[i] = m1[i];
		}
	}

	class op_m2_exch_c implements Operation {
		public void execute(int opcode) {
			int i;
			byte t;
			for (i = 0; i < WSIZE; i++) {
				t = c[i];
				c[i] = m2[i];
				m2[i] = t;
			}
		}
	}

	class op_m2_to_c implements Operation {
		public void execute(int opcode) {
			int i;
			for (i = 0; i < WSIZE; i++)
				c[i] = m2[i];
		}
	}

	class op_f_to_a implements Operation {
		public void execute(int opcode) {
			a[0] = f;
		}
	}

	class op_f_exch_a implements Operation {
		public void execute(int opcode) {
			byte t;
			t = a[0];
			a[0] = f;
			f = t;

		}
	}

	class op_circulate_a_left implements Operation {
		public void execute(int opcode) {

			int i;
			byte t;
			t = a[WSIZE - 1];
			for (i = WSIZE - 1; i >= 1; i--)
				a[i] = a[i - 1];
			a[0] = t;

		}
	}

	class op_bank_switch implements Operation {
		public void execute(int opcode) {

			bank ^= 1;
		}
	}

	class op_c_to_addr implements Operation {
		public void execute(int opcode) {

			ram_addr = (c[1] << 4) + c[0];

			if (ram_addr >= max_ram)
				System.out.println("c -> ram addr: address " + ram_addr
						+ " out of range\n");
		}
	}

	class op_c_to_data implements Operation {
		public void execute(int opcode) {
			int i;
			if (ram_addr >= max_ram) {
				System.out.println("c -> data: address " + ram_addr
						+ " out of range\n");
				return;
			}
			for (i = 0; i < WSIZE; i++)
				ram[ram_addr][i] = c[i];

		}
	}

	class op_data_to_c implements Operation {
		public void execute(int opcode) {
			int i;
			if (ram_addr >= max_ram) {
				System.out.println("c -> data: address " + ram_addr
						+ " out of range, loading 0\n");
				for (i = 0; i < WSIZE; i++)
					c[i] = 0;
				return;
			}
			for (i = 0; i < WSIZE; i++)
				c[i] = ram[ram_addr][i];
		}
	}

	class op_c_to_register implements Operation {
		public void execute(int opcode) {
			int i;

			ram_addr &= ~017;
			ram_addr += (opcode >> 6);

			if (ram_addr >= max_ram) {
				System.out.println("c -> data: address " + ram_addr
						+ " out of range, loading 0\n");
				for (i = 0; i < WSIZE; i++)
					c[i] = 0;
				return;
			}
			for (i = 0; i < WSIZE; i++)
				ram[ram_addr][i] = c[i];
		}
	}

	class op_register_to_c implements Operation {
		public void execute(int opcode) {
			int i;

			ram_addr &= ~017;
			ram_addr += (opcode >> 6);
			// debug("ram_addr = " + Integer.toOctalString(ram_addr));

			if (ram_addr >= max_ram) {
				debug("out of range");
				System.out.println("c -> data: address " + ram_addr
						+ " out of range, loading 0\n");
				for (i = 0; i < WSIZE; i++)
					c[i] = 0;
				return;
			}

			for (i = 0; i < WSIZE; i++)
				c[i] = ram[ram_addr][i];

		}
	}

	class op_clear_data_regs implements Operation {
		public void execute(int opcode) {
			int base;
			int i, j;

			base = ram_addr & ~017;
			for (i = base; i <= base + 15; i++)
				for (j = 0; j < WSIZE; j++)
					ram[i][j] = 0;

		}
	}

	class op_c_to_stack implements Operation {
		public void execute(int opcode) {
			int i;

			for (i = 0; i < WSIZE; i++) {
				t[i] = z[i];
				z[i] = y[i];
				y[i] = c[i];
			}
		}
	}

	class op_stack_to_a implements Operation {
		public void execute(int opcode) {
			int i;
			for (i = 0; i < WSIZE; i++) {
				a[i] = y[i];
				y[i] = z[i];
				z[i] = t[i];
			}
		}
	}

	class op_y_to_a implements Operation {
		public void execute(int opcode) {
			int i;
			for (i = 0; i < WSIZE; i++) {
				a[i] = y[i];
			}
		}
	}

	class op_down_rotate implements Operation {
		public void execute(int opcode) {
			int i;
			byte tp;
			for (i = 0; i < WSIZE; i++) {
				tp = c[i];
				c[i] = y[i];
				y[i] = z[i];
				z[i] = t[i];
				t[i] = tp;
			}
		}
	}

	class op_clear_reg implements Operation {
		public void execute(int opcode) {
			int i;
			for (i = 0; i < WSIZE; i++)
				a[i] = b[i] = c[i] = y[i] = z[i] = t[i] = 0;
			f = 0;
		}
	}

	//

	class op_load_constant implements Operation {
		public void execute(int opcode) {

			if (p >= WSIZE) {

			} else
				c[p] = (byte) (opcode >> 6);

			if (p != 0)
				p--;
			else
				p = WSIZE - 1;

		}
	}

	class op_set_s implements Operation {
		public void execute(int opcode) {
			// debug("entering set_s");
			if ((opcode >> 6) >= SSIZE) {
				System.out.println("stat >= SSIZE at " + prev_pc);
				debug("Fatal .!!!");
			} else
				s[opcode >> 6] = 1;
		}
	}

	class op_clr_s implements Operation {
		public void execute(int opcode) {
			// debug("entering reset_s");

			if ((opcode >> 6) >= SSIZE)
				System.out.println("stat >= SSIZE at " + prev_pc);
			else
				s[opcode >> 6] = 0;
		}
	}

	class op_test_s implements Operation {
		public void execute(int opcode) {
			if ((opcode >> 6) >= SSIZE)
				System.out.println("stat >= SSIZE at " + prev_pc);
			else
				carry = s[opcode >> 6];
		}
	}

	class op_test_s_eq_0 implements Operation {
		public void execute(int opcode) {

			if_flag = true;
			if (s[opcode >> 6] != 0)
				carry = 1;
			else
				carry = 0;
		}
	}

	class op_test_s_eq_1 implements Operation {
		public void execute(int opcode) {

			if_flag = true;
			if (s[opcode >> 6] == 0)
				carry = 1;
			else
				carry = 0;

		}
	}

	static int p_set_map[] = { 14, 4, 7, 8, 11, 2, 10, 12, 1, 3, 13, 6, 0, 9,
			5, 14 };

	static int p_test_map[] = { 4, 8, 12, 2, 9, 1, 6, 3, 1, 13, 5, 0, 11, 10,
			7, 4 };

	class op_set_p implements Operation {
		public void execute(int opcode) {
			p = p_set_map[opcode >> 6];
			if (p >= 14)
				System.out.println("invalid set p, operand encode is "
						+ (opcode > 6));

		}
	}

	class op_test_p_eq implements Operation {

		public void execute(int opcode) {
			//

			val = p_test_map[opcode >> 6];
			if_flag = true;

			if ((val == 0) && (p_change[1] == +1) && (p_change[2] == +1)) {

				if ((p == 0) || (p == 1)) {
					// debug("match p=0 or p=1");
					carry = 0;
				} else {
					// debug ("no match p<>0 or p<>1");
					carry = 1;
				}

			}

			else if ((val == 0) && (p_change[1] == -1) && (p_change[2] == -1)) {
				// debug("test p = 0 after double decrement");

				if ((p == 0) || (p == (WSIZE - 1))) {
					// debug ("match p=0 or p=13");
					carry = 0;
				} else {
					// debug ("no match p<>0 or p<>13");
					carry = 1;
				}
			}

			else {
				if (p == val)
					carry = 0;
				else
					carry = 1;
			}

		}
	}

	class op_test_p_ne implements Operation {
		public void execute(int opcode) {

			val = p_test_map[opcode >> 6];
			if_flag = true;

			if ((val == 0) && (p_change[1] == +1) && (p_change[2] == +1)) {

				if ((p == 0) || (p == 1)) {
					carry = 0;
				} else {
					carry = 1;
				}

			}

			else if ((val == 0) && (p_change[1] == -1) && (p_change[2] == -1)) {
				// debug("test p = 0 after double decrement");

				if ((p == 0) || (p == (WSIZE - 1))) {
					// debug ("match p=0 or p=13");
					carry = 0;
				} else {
					// debug ("no match p<>0 or p<>13");
					carry = 1;
				}
			}

			else {
				if (p == val)
					carry = 0;
				else
					carry = 1;
			}

			carry ^= 1; // complement carry

		}
	}

	class op_sel_rom implements Operation {
		public void execute(int opcode) {
			pc = ((opcode & 01700) << 2) + (pc & 0377);

		}
	}

	class op_del_sel_rom implements Operation {
		public void execute(int opcode) {
			del_rom = opcode >> 6;
			del_rom_flag = true;
		}
	}

	class op_keys_to_rom_addr implements Operation {
		public void execute(int opcode) {

			pc = pc & ~0377;

			if (del_rom_flag) {
				pc = (del_rom << 8) + (pc & 0377);
				del_rom_flag = false;
			}

			debug("key pressed going to :" + Integer.toOctalString(pc));

			Toolkit.getDefaultToolkit().beep();

			if (key_buf < 0) {
				System.out.println("keys->rom address with no key pressed\n");
				return;
			}

			pc += key_buf;
		}
	}

	class op_a_to_rom_addr implements Operation {
		public void execute(int opcode) {

			// debug("Pc : " + Integer.toOctalString( (pc-1 ) ));
			pc = pc & ~0377;
			if (del_rom_flag) {
				pc = (del_rom << 8) + (pc & 0377);
				del_rom_flag = false;
			}
			pc += ((a[2] << 4) + a[1]);

		}
	}

	class op_display_off implements Operation {
		public void execute(int opcode) {
			display_enable = false;
		}
	}

	class op_display_toggle implements Operation {
		public void execute(int opcode) {
			// display_enable = ! display_enable;
			if (display_enable == true)
				display_enable = false;
			else
				display_enable = true;

		}
	}

	// ------------------- added for HP67 -------------------

	class op_keys_to_a implements Operation {
		public void execute(int opcode) {

			if (key_buf < 0) {
				debug("keys->a with no key pressed");
				a[2] = 0;
				a[1] = 0;
				return;
			}

			int temp;
			temp = key_buf >> 4;
			a[2] = (byte) temp; // key_buf >> 4;
			temp = key_buf & 0x0f;
			a[1] = (byte) temp; // key_buf & 0x0f;
			Toolkit.getDefaultToolkit().beep();
		}

	}

	class op_display_reset_twf implements Operation {
		public void execute(int opcode) {

			display_14_digit = true;

		}
	}

	// ----

	class op_model implements Operation {
		public void execute(int opcode) {

		}
	}

	// CARD READER CIRCUIT simulation - 28 12 2007 ///////////////////////

	static String content;

	void CardWrite() {

		int base;
		int i, j;

		byte digit;

		StringBuffer reg = new StringBuffer(14);

		for (i = 7; i < WSIZE; i++)
			reg.append(' ');

		base = 0x99;

		if (record == 0) {

			// Hack for HW header

			for (j = 7; j < WSIZE; j++) {
				digit = 0x00; // ram [base][j];
				reg.insert(WSIZE - j, digit);
			}
		} else {

			for (j = 7; j < WSIZE; j++) {
				digit = ram[base][j];
				reg.insert(WSIZE - j, Character.forDigit(digit, 16));
			}
		}

		magcard(reg.toString());

		record++;

	}

	void CardRead() {

		int i, j, lt;
		int base;
		int num = 0;
		int offset = 0;
		int k = 0;
		char tp1;
		int coderead[] = new int[245];
		// 245 = 35 * 7
		// 7 digits per line
		// 35 lines
		// 1 card 245 char, 525 with spaces + 35 EOL = 560

		content = cardtext.getText();
		lt = content.length();

		for (i = 0; i < lt; i++)

		{
			tp1 = content.charAt(i);

			// filters
			if (tp1 == '\n')
				continue;
			if (tp1 == '\r')
				continue;
			if (tp1 == ' ')
				continue;
			if (tp1 < '0')
				continue;
			if (tp1 > 'f')
				continue;

			num = Character.digit(tp1, 16);

			coderead[k] = num;
			k = k + 1;
		}

		base = 0x9b;

		offset = record * 7;

		if (record == 0) {

			for (j = 7; j < WSIZE; j++)

			{
				ram[base][20 - j] = (byte) coderead[(j + offset) - 7];
				// debug("RAM " + Integer.toString(ram [base][20-j])+ " "
				// +Integer.toString(j));

			}

		}
		if (record == 1) {
			for (j = 7; j < WSIZE; j++)

			{
				ram[base][20 - j] = (byte) coderead[(j + offset) - 7];
				// debug("RAM " + Integer.toString(ram [base][20-j])+ " "
				// +Integer.toString(j));

			}

		}
		if (record == 2) {

			for (j = 7; j < WSIZE; j++)

			{
				ram[base][20 - j] = (byte) coderead[(j + offset) - 7];
				// debug("RAM " + Integer.toString(ram [base][20-j])+ " "
				// +Integer.toString(j));

			}

		}

		if (record > 2) {
			if ((record % 2) == 0) {

				for (j = 7; j < WSIZE; j++)

				{
					ram[base][20 - j] = (byte) coderead[(j + offset) - 7];
					// debug("RAM " + Integer.toString(ram [base][20-j])+ " "
					// +Integer.toString(j));

				}

			} else {

				// debug("REC " + Integer.toString(record));
				for (j = 7; j < WSIZE; j++)

				{
					ram[base][13 - j] = (byte) coderead[(j + offset) - 7];
					// debug("RAM " + Integer.toString(ram [base][20-j])+ " "
					// +Integer.toString(j));
					ram[base][20 - j] = (byte) coderead[(j + offset) - 7];
					// debug("RAM " + Integer.toString(ram [base][20-j])+ " "
					// +Integer.toString(j));

				}

			}
		}

		record++;

	}

	// END OF CARD READER CIRCUIT simulation - 28 12 2007
	// ///////////////////////

	class op_crc_1500 implements Operation {
		public void execute(int opcode) {

			if (pause == true) {
				s[3] = 1;
				pause = false;
			} else
				s[3] = 0;

		}
	}

	class op_crc_1400 implements Operation {
		public void execute(int opcode) {

			pause = true;

		}
	}

	class op_crc_300 implements Operation {
		public void execute(int opcode) {

			// Test the RUN/RPGM switch

			if (prgm_mode == true)
				s[3] = 1;
			else
				s[3] = 0;

		}
	}

	class op_crc_500 implements Operation {
		public void execute(int opcode) {

			if (a_key == true) {
				s[3] = 1;
				a_key = false;
			}

			else
				s[3] = 0;

		}
	}

	class op_crc_400 implements Operation {
		public void execute(int opcode) {

			a_key = true;

		}
	}

	class op_crc_100 implements Operation {
		public void execute(int opcode) {

			// Device always ready

			s[3] = 1;
			card_inserted = 0;

		}
	}

	class op_crc_1700 implements Operation {
		public void execute(int opcode) {

			// Read Write to Card

			if (prgm_mode == false)

			{
				// RUN MODE

				if ((read_card == 0) && (write_card == 1)) // WRITE IN RUN MODE
															// "f W/DATA"
				{
					// debug("WRITE RUN" + Integer.toString(block) + " " +
					// Integer.toOctalString(pc-1) );

					if (!(block == 0 | block == 2 | block == 37)) { // skipped
						// debug("write#  " + Integer.toString(block));
						CardWrite();

						s[3] = 0;
						read_card = 0;
					}

					block++;
					if (block == 38)
						block = 0; // # of calls 1700 per card
				}

				if (read_card == 1) // READ IN RUN MODE DEFAULT
				{
					// debug("READ RUN" + Integer.toString(block) + " " +
					// Integer.toOctalString(pc-1) );

					if (!(block == 0 | block == 36 | block == 37)) {
						// debug("read# " + Integer.toString(block));
						CardRead();

						s[3] = 0;
						def_fn = false;
					}

					block++;
					if (block == 38)
						block = 0; // total calls per card

				}

			}

			else { // PRGM mode : Write PRGM

				// debug("op1700 write" + Integer.toString(block) + " " +
				// Integer.toOctalString(pc-1) );

				if (!(block == 0 | block == 2 | block == 37)) { // skipped
					// debug("write# " + Integer.toString(block));
					CardWrite();

					s[3] = 0;
					read_card = 0;
				}
				block++;
				if (block >= 38)
					block = 0;

			}

		}

	}

	class op_crc_160 implements Operation {
		public void execute(int opcode) {

			if (display_digits == true) {
				s[3] = 1;
				display_digits = false;
			} else
				s[3] = 0;
		}
	}

	class op_crc_60 implements Operation {
		public void execute(int opcode) {

			display_digits = true;
		}
	}

	class op_crc_560 implements Operation {
		public void execute(int opcode) {

			if (card_inserted == 1)
				s[3] = 1;
			else
				s[3] = 0;

		}
	}

	class op_crc_1300 implements Operation {
		public void execute(int opcode) {

			if (merge == true) {

				s[3] = 1;
				merge = false;
			} else
				s[3] = 0;

		}
	}

	class op_crc_1200 implements Operation {
		public void execute(int opcode) {

			merge = true;

		}
	}

	class op_crc_360 implements Operation {
		public void execute(int opcode) {
			// motor off
			// CRCmessage("op360");
			record = 0; // motor off resets record pointer
			block = 0;
			if (read_card == 1)
				cardtext.setText("");
		}
	}

	class op_crc_260 implements Operation {
		public void execute(int opcode) {
			// CRCmessage("op260");
			// motor on
		}
	}

	class op_crc_660 implements Operation {
		public void execute(int opcode) {
			// CRCmessage("op660");

			cardtext.setText("");
			// write mode

			write_card = 1; // write mode data or rpgm
			read_card = 0;

		}
	}

	class op_crc_760 implements Operation {
		public void execute(int opcode) {
			// set read mode

			magcardRaw("");
			read_card = 1;
		}
	}

	class op_crc_1000 implements Operation {
		public void execute(int opcode) {

			def_fn = true;
		}
	}

	class op_crc_1100 implements Operation {
		public void execute(int opcode) {

			// CRCmessage("op1100");

			if (def_fn == true) {
				s[3] = 1;
				def_fn = false;
			} else
				s[3] = 0;

		}
	}

	Operation op_fcn[] = new Operation[1024];

	void init_ops() {
		int i;

		for (i = 0; i < 1024; i += 4) {
			op_fcn[i + 0] = new bad_op();
			op_fcn[i + 1] = new op_jsb(); /* type 1: aaaaaaaa01 */
			op_fcn[i + 2] = new op_arith(); /* type 2: ooooowww10 */
			op_fcn[i + 3] = new op_goto(); /* type 1: aaaaaaaa11 */
		}

		for (i = 0; i <= 15; i++) {
			/* xx00 uassigned */
			op_fcn[00004 + (i << 6)] = new op_set_s();
			/* xx10 misc */
			op_fcn[00014 + (i << 6)] = new op_clr_s();
			/* xx20 misc */
			op_fcn[00024 + (i << 6)] = new op_test_s_eq_1();
			op_fcn[00030 + (i << 6)] = new op_load_constant();
			;
			op_fcn[00034 + (i << 6)] = new op_test_s_eq_0();
			;
			op_fcn[00040 + (i << 6)] = new op_sel_rom();
			op_fcn[00044 + (i << 6)] = new op_test_p_eq();
			op_fcn[00050 + (i << 6)] = new op_c_to_register();
			op_fcn[00054 + (i << 6)] = new op_test_p_ne();
			/* xx60 misc */
			op_fcn[00064 + (i << 6)] = new op_del_sel_rom();
			op_fcn[00070 + (i << 6)] = new op_register_to_c();
			op_fcn[00074 + (i << 6)] = new op_set_p();
		}

		op_fcn[00000] = new op_nop();
		op_fcn[00070] = new op_data_to_c();

		op_fcn[00010] = do_op_clear_reg = new op_clear_reg();
		op_fcn[00110] = do_op_clear_s = new op_clear_s();
		op_fcn[00210] = new op_display_toggle();
		op_fcn[00310] = new op_display_off();
		op_fcn[00410] = new op_m1_exch_c();
		op_fcn[00510] = new op_m1_to_c();
		op_fcn[00610] = new op_m2_exch_c();
		op_fcn[00710] = new op_m2_to_c();
		op_fcn[01010] = new op_stack_to_a();
		op_fcn[01110] = new op_down_rotate();
		op_fcn[01210] = new op_y_to_a();
		op_fcn[01310] = new op_c_to_stack();
		op_fcn[01410] = new op_decimal();
		/* 1510 unassigned */
		op_fcn[01610] = new op_f_to_a();
		op_fcn[01710] = new op_f_exch_a();

		op_fcn[00020] = new op_keys_to_rom_addr();
		op_fcn[00120] = new op_keys_to_a();
		op_fcn[00220] = new op_a_to_rom_addr();
		op_fcn[00320] = new op_display_reset_twf();
		op_fcn[00420] = new op_binary();
		op_fcn[00520] = new op_circulate_a_left();
		op_fcn[00620] = new op_dec_p();
		op_fcn[00720] = new op_inc_p();
		op_fcn[01020] = new op_return();
		/* 1120..1720 unknown, probably printer */

		/* 0160..0760 unassigned/unknown */

		op_fcn[01060] = new op_bank_switch();
		op_fcn[01160] = new op_c_to_addr();
		op_fcn[01260] = new op_clear_data_regs();
		op_fcn[01360] = new op_c_to_data();

		
		
		op_fcn[01760] = new op_nop(); /* "HI I'M WOODSTOCK" */

		// CRC chip in 67/97
		op_fcn[00100] = new op_crc_100(); 
		op_fcn[01500] = new op_crc_1500(); 

		op_fcn[00300] = new op_crc_300();
		op_fcn[00400] = new op_crc_400(); 
		op_fcn[00500] = new op_crc_500(); 
		op_fcn[01000] = new op_crc_1000(); 
		op_fcn[01100] = new op_crc_1100(); 
		op_fcn[01200] = new op_crc_1200(); 
		op_fcn[01300] = new op_crc_1300();
		op_fcn[01400] = new op_crc_1400(); 
		op_fcn[01700] = new op_crc_1700(); 

		op_fcn[00060] = new op_crc_60(); 
		op_fcn[00160] = new op_crc_160(); 
		op_fcn[00260] = new op_crc_260(); 
		op_fcn[00360] = new op_crc_360(); 
		op_fcn[00560] = new op_crc_560(); 
		op_fcn[00660] = new op_crc_660(); 
		op_fcn[00760] = new op_crc_760(); 

	}

	String pc_to_octal(int pc) {
		String t = Integer.toOctalString(p);
		if (t.length() == 1)
			return "00" + t;
		else if (t.length() == 2)
			return "0" + t;
		else
			return t;
	}

	void disassemble_instruction(int g, int r, int p, int opcode) {
		int i;
		System.out.print("L" + Integer.toOctalString(g)
				+ Integer.toOctalString(r) + pc_to_octal(p));
		for (i = 0x200; i > 0; i >>= 1)
			System.out.print(((opcode & i) != 0) ? "1" : ".");
	}

	// ----------------------------------------------------------------

	void init_breakpoints() {
		int g, r, p;

		for (g = 0; g < MAX_GROUP; g++)
			for (r = 0; r < MAX_ROM; r++)
				for (p = 0; p < ROM_SIZE; p++)
					bpt[g][r][p] = 1;
	}

	void init_source() {
		int g, r, p;

		for (g = 0; g < MAX_GROUP; g++)
			for (r = 0; r < MAX_ROM; r++)
				for (p = 0; p < ROM_SIZE; p++)
					source[g][r][p] = null;
	}

	int gt, rt, pt, opcodet;

	boolean parse_address(String oct) {
		String s;

		try {
			// No Group, no ROM
			s = oct.substring(0, 1);
			gt = Integer.parseInt(s, 8);
			gt = 0;
			s = oct.substring(1, 2);
			rt = Integer.parseInt(s, 8);
			rt = 0;

			s = oct.substring(0, 5);
			pt = Integer.parseInt(s, 8);
			return true;
		} catch (NumberFormatException e) {
			return false;
		}
	}

	boolean parse_opcode(String bin) {
		int i;

		opcodet = 0;
		for (i = 0; i < 10; i++) {
			opcodet <<= 1;
			if (bin.charAt(i) == '1')
				opcodet += 1;
			else if (bin.charAt(i) == '.')
				opcodet += 0;
			else
				return false;
		}
		return true;
	}

	String trimSource(String str) {
		String t;

		if (str.length() < 48)
			return new String(str);

		if (str.length() < 80)
			t = str;
		else
			t = str.substring(0, 80);

		return t.substring(8, 15) + t.substring(16, 27) + t.substring(28, 31)
				+ t.substring(32, 38) + t.substring(48);

	}

	void read_listing_file(String fn) throws IOException {

		String buf;
		int count = 0;

		do_trace("Reading microcode file in Jar");
		try {

			InputStream isr = getClass().getResourceAsStream(fn);
			BufferedReader br = new BufferedReader(new InputStreamReader(isr));

			while ((buf = br.readLine()) != null) {
				if ((buf.length() >= 25) && (buf.charAt(7) == 'L')
						&& (buf.charAt(13) == ':')
						&& parse_address(buf.substring(8))
						&& parse_opcode(buf.substring(16))) {
					if ((gt >= MAX_GROUP) || (rt >= MAX_ROM)
							|| (pt >= ROM_SIZE))
						System.out.println("bad address");
					else if (bpt[gt][rt][pt] == 0) {
						System.out
								.println("duplicate listing line for address "
										+ Integer.toOctalString(gt)
										+ Integer.toOctalString(rt)
										+ pc_to_octal(pt));
						System.out.println("orig: " + source[gt][rt][pt]);
						System.out.println("dup:  " + buf);
					} else {

						ucode[gt][rt][pt] = (short) opcodet;

						bpt[gt][rt][pt] = 0;
						if (keep_src)
							source[gt][rt][pt] = trimSource(buf);
						count++;
					}
				}
			}
			System.out.println("Read " + count + " words from " + fn);
			do_trace("Read " + count + " words of microcode from ROM image");
		} catch (MalformedURLException e) {
			fatal("Can't access firmware file");
		}
	}

	String reg_to_string(byte r[]) {
		int i;
		String st = new String("");
		String tp = new String("");

		for (i = WSIZE - 1; i >= 0; i--)

		{
			if (r[i] == 15)
				tp = " f";
			else
				tp = Integer.toHexString(r[i]);
			st += tp;
		}
		return st;
	}

	// Convert Status
	String stat_to_string() {
		int i;
		String st = new String("");
		for (i = 0; i < SSIZE; i++)
			st += (s[i] == 1) ? Integer.toHexString(i) : ".";
		return st;
	}

	// Reg 14
	String Reg14_to_string() {
		int i;
		String st = new String("");
		for (i = 0; i < SSIZE; i++)
			st += (s[i] == 1) ? Integer.toHexString(i) : ".";
		return st;
	}

	public static char CharacterforDigit(int digit) {

		char ttp;
		ttp = Character.forDigit(digit, 16);
		if (ttp == '\0')
			ttp = ' ';
		if (ttp == 'f')
			ttp = ' ';
		if (ttp == 'a')
			ttp = 'r';
		if (ttp == 'b')
			ttp = 'C';
		if (ttp == 'c')
			ttp = 'o';
		if (ttp == 'd')
			ttp = 'd';
		if (ttp == 'e')
			ttp = 'E';

		return ttp;
	}

	// /////////////////////////////// Display ///////////////////
	void handle_io() {
		String dstr = new String("");
		StringBuffer dis = new StringBuffer(15);
		char tp;

		int i, k, d, ee;

		ee = 15;

		for (i = 0; i < ee; i++)
			dis.append(' ');

		if (display_enable == true) {

			if (prgm_mode == false) {

				// exponent EE

				if (b[0] == 0) {

					tp = CharacterforDigit(a[0]);
					dis.setCharAt(ee - 1, tp);
				} else
					dis.setCharAt(ee - 1, ' ');

				if ((b[1] == 0) || (b[1] == 4)) {

					tp = CharacterforDigit(a[1]);
					dis.setCharAt(ee - 2, tp);
				} else
					dis.setCharAt(ee - 2, ' ');

				// signs
				// 2

				if (a[2] == 15) {
					dis.setCharAt(0, ' '); // PGM mode no sign
					dis.setCharAt(ee - 3, ' '); // before step # or key code
				}

				if (a[2] == 0) {
					dis.setCharAt(0, '-'); // matissa -
					dis.setCharAt(ee - 3, ' '); // exponent +
				}

				if (a[2] == 1) {
					dis.setCharAt(0, ' '); // mantissa +
					dis.setCharAt(ee - 3, ' '); // exponent +
				}

				if ((a[2] == 2) || (a[2] == 8)) {
					dis.setCharAt(0, '-'); // matissa -
					dis.setCharAt(ee - 3, '-'); // exponent -
				}

				if ((a[2] == 3) || (a[2] == 9)) {
					dis.setCharAt(0, ' '); // matissa +
					dis.setCharAt(ee - 3, '-'); // exponent -
				}

				// mantissa
				// 3
				k = 3;
				d = 11;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 4
				k = 4;
				d = 10;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 5
				k = 5;
				d = 9;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 6
				k = 6;
				d = 8;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 7
				k = 7;
				d = 7;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 8
				k = 8;
				d = 6;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 9
				k = 9;
				d = 5;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 10
				k = 10;
				d = 4;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 11
				k = 11;
				d = 3;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 12
				k = 12;
				d = 2;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

				// 13
				k = 13;
				d = 1;
				if (b[k] == 0) {
					tp = CharacterforDigit(a[k]);
					dis.setCharAt(d, tp);
				}
				if (b[k] == 3)
					dis.setCharAt(d, '.');

			} // end of run mode display

			else {

				// PGM display

				k = 0;
				d = 13;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 1;
				d = 12;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 2;
				d = 11;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, ' ');

				k = 3;
				d = 10;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 4;
				d = 9;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 5;
				d = 8;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, ' ');

				k = 6;
				d = 7;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 7;
				d = 6;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 8;
				d = 5;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, ' ');

				k = 9;
				d = 4;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 10;
				d = 3;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 11;
				d = 2;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 12;
				d = 1;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

				k = 13;
				d = 0;
				tp = CharacterforDigit(a[k]);
				dis.setCharAt(d, tp);

			}

			dstr = dis.toString();

		}

		update_display(dstr);
		i = check_keyboard();

		if (i == PGM_MODE)

		{
			i = -1;
			Toolkit.getDefaultToolkit().beep();
			prgm_mode = true;

		}

		if (i == RUN_MODE)

		{
			i = -1;
			Toolkit.getDefaultToolkit().beep();
			prgm_mode = false;

		}

		if (i == OFF_MODE) {
			i = -1;
			Toolkit.getDefaultToolkit().beep();
			dkeys[RUN].value = false;
			update_run_key();
		}

		if (i == ON_MODE) {
			i = -1;
			Toolkit.getDefaultToolkit().beep();
			dkeys[RUN].value = true;
			// restarted = true;
			anyBPsEnabled = dkeys[ENABLE1].value || dkeys[ENABLE2].value
					|| dkeys[ENABLE3].value;
			update_run_key();
		}

		pressed_key = -1;
		if (i >= 0) {
			key_flag = true;
			key_buf = i;
		} else
			key_flag = false;

		if ((i >= 0) && (black == true)) {

			gr.drawImage(img, 0, 0, this);
			black = false;
		}

		if ((key_flag) && (i == 0020)) {
			black = true;
			gr.drawImage(himg, 0, 0, this);
		}

	}

	//
	// Debug console code follows
	//

	class CircAddressBuffer {
		int size;
		String buf[];
		int curPos;

		public CircAddressBuffer(int size) {
			this.size = size;
			buf = new String[size];
			curPos = size - 1;
			for (int i = 0; i < size; i++)
				buf[i] = "";
		}

		public void put(String s) {
			buf[curPos--] = new String(s);
			if (curPos < 0)
				curPos = size - 1;
		}

		
		public String get(int position) {
			int p;

			p = curPos + position;
			if (p > size - 1)
				p -= size;
			return new String(buf[p]);
		}

	}

	//
	final static int MAXTRACE = 4;
	
	final static int MAXTRACEPAGES = 1999;
	
	final static int TRACEHEIGHT = 19;
	
	final static int TRACEFONTSIZE = 13; 
	
	final static int TRACEWIDTH = 889;
	
	final static String TRACETESTSTRING = new String(
			"01234567890123456789012345678901234567890123456789012345678901");

	Rectangle tPageRect;
	Rectangle traceRect[] = new Rectangle[MAXTRACE];

	final static int MAXNEXT = 5;
	Rectangle nextRect[] = new Rectangle[MAXNEXT];
	String nextString[] = new String[MAXNEXT];

	CircAddressBuffer trace_buf;
	int tracePage = 0;
	int BpIn = 0;

	void init_trace() {
		int i = TRACEFONTSIZE;

		// reduce font size if necessary to fit the debugger window

		for (i = TRACEFONTSIZE; i > 6; i--) {
			trace_font = new Font("Courier", Font.PLAIN, i);
			tfm = getFontMetrics(trace_font);

			if ((TRACEWIDTH >= tfm.stringWidth(TRACETESTSTRING) + 2)
					&& (TRACEHEIGHT >= (tfm.getHeight() + 2)))
				break;
		}

		// BP input zone
		tPageRect = new Rectangle(540, 203, 60, TRACEHEIGHT);
		//

		for (i = 0; i < MAXTRACE; i++) {
			traceRect[i] = new Rectangle(265, 300 + i * TRACEHEIGHT,
					TRACEWIDTH, TRACEHEIGHT);

		}
		;
		trace_buf = new CircAddressBuffer(MAXTRACE * MAXTRACEPAGES);
	}

	void init_next() {
		int i;

		for (i = 0; i < MAXNEXT; i++) {
			nextRect[i] = new Rectangle(265, 394 + i * TRACEHEIGHT, TRACEWIDTH,
					TRACEHEIGHT);
			nextString[i] = "";
		}
		;
	}

	void add_trace(String s) {
		trace_buf.put(s);
	}

	synchronized void display_trace() { // Draw trace zone above 'PC'
		int i;

		gr.setFont(trace_font);
		for (i = 1; i <= MAXTRACE; i++) {
			draw_string_left(gr, trace_buf.get(i + tracePage * MAXTRACE),
					traceRect[MAXTRACE - i], BLACK, WHITE, tfm);
		}
		draw_string_left(gr, Integer.toString(BpIn), tPageRect, BLACK, WHITE,
				tfm);
		gr.setFont(key_font);
	}

	void do_trace(String str) { // display trace zone above 'PC'
		tracePage = 0;
		add_trace(str);
		debug(str);
		display_trace();
	}

	void add_next(String str) {
		int i;

		for (i = 1; i < MAXNEXT; i++) {
			nextString[i - 1] = nextString[i];
		}

		nextString[MAXNEXT - 1] = new String(str);

	}

	synchronized void display_next() { // draw 'Next' box below 'PC'
		int i;

		gr.setFont(trace_font);
		for (i = 0; i < MAXNEXT; i++) {
			draw_string_left(gr, nextString[i], nextRect[i], BLACK, WHITE, tfm);
		}
		gr.setFont(key_font);
	}

	void do_next(String str) {
		add_next(str);
		display_next(); // zone below 'PC'
	}

	final static int MAXBPS = 3;
	Rectangle bpRect[] = new Rectangle[MAXBPS];
	int bpReg[] = new int[MAXBPS];
	int card_inserted = 0;
	int read_card = 0;
	int write_card = 0;

	int bpinput = 0; // keyboard input for user breakpoints

	boolean anyBPsEnabled = false;

	void init_user_bp_display() {
		int i;

		for (i = 0; i < MAXBPS; i++) {
			bpRect[i] = new Rectangle(600, 28 + i * 30, 60, TRACEHEIGHT);
			bpReg[i] = 0;
		}

	}

	void display_user_bps() {
		int i;

		for (i = 0; i < MAXBPS; i++) {
			draw_string_left(gr, Integer.toOctalString(bpReg[i]), bpRect[i],
					BR_RED, DK_RED, tfm);
		}
	}

	void display_bpzone() {
		draw_string_left(gr, Integer.toString(BpIn), tPageRect, BLACK, WHITE,
				tfm);
	}

	void display_mem_pgm() {
		int i;
		int j;

		ram_addr = 057;

		for (j = ram_addr; j > 017; j--) {
			debug("addr = " + Integer.toOctalString(j) + " (0->13) ");
			{
				for (i = 0; i < WSIZE; i++)
					debugRaw(Integer.toHexString(ram[j][i]) + " ");
			}
			debugRaw("\r\n");
		}

	}

	void display_mem_data(Integer reg) {
		int i;
		int j;

		if (reg == 0) {

			ram_addr = 077;

			for (j = ram_addr; j > 057; j--) {
				debug("addr = " + Integer.toOctalString(j) + " (0->13) ");
				{
					for (i = 0; i < WSIZE; i++)
						debugRaw(Integer.toString(ram[j][i]) + " ");
				}
				debugRaw("\r\n");
			}

			ram_addr = 017;

			for (j = ram_addr; j >= 0; j--) {
				debug("addr = " + Integer.toOctalString(j) + " (0->13) ");
				{
					for (i = 0; i < WSIZE; i++)
						debugRaw(Integer.toString(ram[j][i]) + " ");
				}
				debugRaw("\r\n");
			}
		}

		else if (reg < 78) {

			debug("addr = " + Integer.toOctalString(reg) + " (0->13) ");
			for (i = 0; i < WSIZE; i++)
				debugRaw(Integer.toString(ram[reg][i]) + " ");
			debugRaw("\r\n");

		}
	}

	void display_status_data() {

		int i;
		debug("Status Reg14 addr" + Integer.toOctalString(076) + " (0->13) ");
		for (i = 0; i < WSIZE; i++)
			debugRaw(Integer.toString(ram[076][i]) + " ");
		debugRaw("\r\n");

	}

	final static int MAXREGS = 11; // 9;
	Rectangle regRect[] = new Rectangle[MAXREGS];
	Rectangle regLabelRect[] = new Rectangle[MAXREGS];
	String prevReg[] = new String[MAXREGS];
	final static int DISPLAYREGWIDTH = 116;

	void init_reg_display() {
		int i;

		for (i = 0; i < MAXREGS; i++) {
			regRect[i] = new Rectangle(382, 25 + i * TRACEHEIGHT,
					DISPLAYREGWIDTH, TRACEHEIGHT);
			regLabelRect[i] = new Rectangle(362, 25 + i * TRACEHEIGHT, 20,
					TRACEHEIGHT);
			prevReg[i] = new String("");
		}

	}

	void maybe_display(String s, boolean repaint, int reg) {
		if (repaint || !prevReg[reg].equals(s)) {
			draw_string_left(gr, " " + s, regRect[reg], BLACK, WHITE, tfm);
			prevReg[reg] = s;
		}
	}

	void display_regs(boolean repaint) {
		int i = 0;
		// String temp;
		if (dkeys[DEBUG].value) {
			if (repaint) {
				draw_string(gr, "A:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "B:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "C:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "D:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "E:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "F:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "M1:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "M2:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "F:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "P:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
				draw_string(gr, "St:", regLabelRect[i++], BLACK, VLT_GREY, kfm);
			}

			i = 0;
			maybe_display(reg_to_string(a), repaint, i);
			i++;
			maybe_display(reg_to_string(b), repaint, i);
			i++;
			maybe_display(reg_to_string(c), repaint, i);
			i++;
			maybe_display(reg_to_string(y), repaint, i);
			i++;
			maybe_display(reg_to_string(z), repaint, i);
			i++;
			maybe_display(reg_to_string(t), repaint, i);
			i++;
			maybe_display(reg_to_string(m1), repaint, i);
			i++;
			maybe_display(reg_to_string(m2), repaint, i);
			i++;

			maybe_display(Byte.toString(f), repaint, i);
			i++;
			maybe_display(Integer.toHexString(p), repaint, i);
			i++;
			maybe_display(stat_to_string(), repaint, i);
		}
	}

	void my_sleep(long t) {
		try {
			java.lang.Thread.sleep(t); // half sec
		} catch (InterruptedException e) {
			// ignore it
		}
	}

	// display "next PC zone (below 'PC')"-----------------------------------
	void do_nextDisplay(int group, int rom, int tpc) {
		int j;
		for (j = 0; j < MAXNEXT; j++) {

			if (source[group][rom][tpc] != null)
				// debug("NEXT: " + (source [group][rom][tpc]));
				do_next(source[group][rom][tpc]);
			tpc++;
		}
	}

	// --------------------------------------------------------

	public void makedebug() {
		// if (this.getParameter("diag") == null) return;
		debugframe = new Frame("DEBUG ^a, ^c, ^v ^x");

		debugframe.setVisible(true);
		debugframe.setSize(250, 260); // width heigth
	
		debugframe.setLocation(875, 180);
		debugtext = new TextArea();
		debugtext.setFont(new Font("Courier", Font.PLAIN, 14));
		debugframe.add(debugtext);
	}

	public void debug(String s) {
		if (debugframe == null)
			return;
		debugtext.append(s + "\r\n");
		debugframe.setVisible(true);
	}

	public void debugRaw(String str) {
		if (debugframe == null)
			return;
		debugtext.append(str);
		debugframe.setVisible(true);
	}

	public void makecard() {
		
		cardframe = new Frame("CARD ^a, ^c, ^v ^x");

		cardframe.setVisible(true);
		cardframe.setSize(250, 260);
		// Location(X,Y)
		cardframe.setLocation(875, 460); 
		cardtext = new TextArea();
		cardtext.setFont(new Font("Courier", Font.PLAIN, 14));
		cardframe.add(cardtext);
	}

	public void magcard(String s) {

		if (cardframe == null)
			return;
		cardtext.append(s + "\r\n");
		cardframe.setVisible(true);
	}

	public void magcardRaw(String s) {

		if (cardframe == null)
			return;
		cardframe.setVisible(true);
	}

	void emulator() {
		int opcode = 0;
		int sleep_count = 0;
		boolean after_brk = false;

		makedebug();
		makecard();

		// Reset Woodstock

		arithmetic_base = 10;
		pc = 0;
		del_rom_flag = false;
		if_flag = false;
		sp = 0;
		group = 0;
		del_rom = 0;
		del_grp = 0;

		do_op_clear_reg.execute(0);
		do_op_clear_s.execute(0);

		p = 0;
		p_change[2] = 0;
		p_change[1] = 0;
		p_change[0] = 0;

		io_count = 0;
		display_enable = false;
		display_14_digit = false;
		display_digit_position = 0;
		display_scan_position = 0;

		key_buf = -1;
		key_flag = false;
		prgm_mode = false;

		for (;;) {

			// Restart the calc while the Applet is still running

			if (!dkeys[RUN].value) { // RUN button off


				if (after_brk == false) { // Zap display when OFF but NOT after after a break

					// Blank the display
					// and ZAP ACT context
					
					
					display_enable = false;
					int i;
					for (i = 0; i <= 13; i++)
						a[i] = 0; 
					for (i = 0; i <= 13; i++)
						b[i] = 0;
					for (i = 0; i < WSIZE; i++)
						c[i] = y[i] = z[i] = t[i] = 0;
					f = 0;
					p = 0;
					handle_io();
					display_regs(false); // repaint regs

					def_fn = true; // default keys enable

					pc = 0;

				}
				
			}

			while (dkeys[RUN].value || step) {
				
				// RUN ON or STEP ON
				
				if (dkeys[RUN].value)
					after_brk = false;

				if (dkeys[SLOW].value && !step) {
					// slow on and step off

					my_sleep(200);
					check_debug_keyboard(); // need fast way to exit slow mode
				}

				// Make keyboard more responsive in stepping case
				if (step)
					handle_io();

				// display trace window
				if (dkeys[TRACE].value) {
					if (source[group][rom][(bank << 12) + pc] != null) {
						do_trace(source[group][rom][(bank << 12) + pc]);

					} else
						disassemble_instruction(group, rom, (bank << 12) + pc,
								opcode);
				}

				// Execute
				prev_pc = pc;
				opcode = ucode[group][rom][(bank << 12) + pc];

				prev_if_flag = if_flag;
				if_flag = false;

				prev_carry = carry;
				carry = 0;

				p_change[2] = p_change[1];
				p_change[1] = p_change[0];
				p_change[0] = 0;

				if ((pc < 02000) && (bank == 1)) {
					pc = (pc - 1);
					// debug("implicit bank 0 switch at , " +
					// Integer.toOctalString(pc) );
					bank = 0;
				}

				if (key_flag) {
					s[15] = 1; // a key has been pressed
				}

				pc = (pc + 1) & 07777;

				// Execute the "then" branch

				if (prev_if_flag) {
					if (prev_carry == 0) {
						pc = (pc & ~01777);
						pc = pc | opcode;
					}

				}

				// else execute opcode fetched
				else
					op_fcn[opcode].execute(opcode);

				// repaint
				if (dkeys[TRACE].value || dkeys[SLOW].value || step) {
					display_regs(false);

				}

				if (dkeys[TRACE].value || dkeys[SLOW].value || step)

					if (source[group][rom][(bank << 12) + pc] != null) { // below
																			// PC
						// debug("NEXT: " + Integer.toOctalString((bank<<12)
						// +pc)+"\n");
						do_nextDisplay(group, rom, (bank << 12) + pc);

					}

				if (dkeys[TRACE].value && dkeys[REGS].value) {
					do_trace("  A=" + reg_to_string(a) + "  B="
							+ reg_to_string(b) + "  C=" + reg_to_string(c));
					do_trace("  D=" + reg_to_string(y) + "  M1="
							+ reg_to_string(m1) + "  M2=" + reg_to_string(m2));
					do_trace("  F=" + Byte.toString(f) + "  P="
							+ Integer.toHexString(p) + "  S="
							+ stat_to_string());
				}

				step = false;

				io_count--;
				if (io_count <= 0) {
					handle_io();

					// make keyboard more responsive in slow
					// case - handle_io slow down full speed runs
					if (dkeys[SLOW].value)
						io_count = 5;
					else
						io_count = 35;

					if (sleep_count++ > 6) {
						sleep_count = 0;
						my_sleep(71); // drop CPU utilization and make near real
										// time
						display_regs(false); // keep display fresh
					}
				}
				if (anyBPsEnabled) {
					int address = pc + group;
					int i;
					// debug("BP: " + Integer.toOctalString((bank<<12) + address
					// )+ "\n");
					for (i = 0; i < MAXBPS; i++) {
						if (((bank << 12) + address == bpReg[i])
								&& dkeys[ENABLE1 + i].value) {
							dkeys[RUN].value = false;
							after_brk = true;
							update_run_key();
							do_nextDisplay(group, rom, (bank << 12) + pc);
						}

					}

				}
			} // end while

			my_sleep(100); // don't spin if not running
			// calling handle_io() here would lose keystroke in stepping case
			check_debug_keyboard();
		} // end of for(;;) Loop
	} // end if emulator method

	boolean calcRunning = false;

	public void start() {
		if (calcRunning) {
			System.out.println("calc_thread started");
			// calc_thread.resume(); // warning about navigator deadlock
		}
	}

	public void stop() {
		if (calcRunning) {
			System.out.println("calc_thread stoped");
			// calc_thread.suspend(); // warning about navigator deadlock
		}
	}

	class run_calc implements Runnable {

		void get_source_and_run() throws IOException {
			init_breakpoints();
			init_source();
			read_listing_file("hp67.rom");
			init_ops();
			do_trace("HP-67 V2.6 2013 12/02 Ready.  Press run, slow or step.");
			System.out.println("entering the emulator loop");
			emulator();
		}

		public void run() {

			try {
				get_source_and_run();
			} catch (IOException e) {
				// maybe it's just a temporary network problem
				// let's try to get it again
				do_trace("Error reading firmware listing file - retrying");
				System.out
						.println("Error reading firmware listing file - retrying");
				my_sleep(300);
				try {
					get_source_and_run();
				} catch (IOException e2) {
					do_trace("Fatal error reading firmware listing file - 2nd try");
					fatal("Error reading listing firmware file - 2nd try");
				}
			}

		} // end of run
	} // end of run_calc
} // end of sim67